package org.processmining.dataawarereplayer.precision;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.dataawarereplayer.precision.PrecisionConfigPanel.PrecisionType;
import org.processmining.dataawarereplayer.precision.measure.ActivityPrecisionMeasureImpl;
import org.processmining.dataawarereplayer.precision.measure.PrecisionMeasure;
import org.processmining.dataawarereplayer.precision.measure.PrecisionMeasure.ResultMode;
import org.processmining.dataawarereplayer.precision.measure.ResourcePrecisionMeasureImpl;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjectionException;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjector;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjectorFittingImpl;
import org.processmining.dataawarereplayer.precision.projection.ProjectedTrace;
import org.processmining.dataawarereplayer.precision.state.PrecisionState;
import org.processmining.dataawarereplayer.precision.state.StateFactory;
import org.processmining.dataawarereplayer.precision.state.StateFactoryDPNImpl;
import org.processmining.dataawarereplayer.precision.state.StateTransition;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.collect.SetMultimap;

public class DataAwarePrecision {

	private final static class PrecisionResultImpl implements PrecisionResult {

		private final PrecisionMeasure precisionMeasure;

		public PrecisionResultImpl(PrecisionMeasure precisionMeasure) {
			this.precisionMeasure = precisionMeasure;
		}

		public double getPrecision() {
			return precisionMeasure.getPrecision();
		}

		public double getPrecision(Place p) {
			return precisionMeasure.getPrecision(p);
		}

		public String toString() {
			return precisionMeasure.toString();
		}

		public double getPrecision(Marking marking) {
			return precisionMeasure.getPrecision(marking);
		}

		public long getObservedContinuations(Place p) {
			return precisionMeasure.getObservedBehavior(p);
		}

		public long getPossibleContinuations(Place p) {
			return precisionMeasure.getPossibleBehavior(p);
		}

		public long getFrequency(PrecisionState state) {
			return precisionMeasure.getFrequency(state);
		}

		public SetMultimap<PrecisionState, StateTransition> getObservedStateTransitions(Place p) {
			return precisionMeasure.getObservedStateTransitions(p);
		}

		public SetMultimap<PrecisionState, StateTransition> getPossibleStateTransitions(Place p) {
			return precisionMeasure.getPossibleStateTransitions(p);
		}

		public SetMultimap<PrecisionState, StateTransition> getPossibleStateTransitions() {
			return precisionMeasure.getPossibleStateTransitions();
		}

		public SetMultimap<PrecisionState, StateTransition> getObservedStateTransitions() {
			return precisionMeasure.getObservedStateTransitions();
		}

		public long getObservedLocalContinuations(Place p) {
			return precisionMeasure.getObservedLocalBehavior(p);
		}

		public long getPossibleLocalContinuations(Place p) {
			return precisionMeasure.getPossibleLocalBehavior(p);
		}

		public double getLocalPrecision(Place p) {
			return precisionMeasure.getLocalPrecision(p);
		}

		public Object getModel() {
			return precisionMeasure.getModel();
		}

	}

	private final static class TracesProcessor implements Callable<Void> {
		private final ProjectedTrace processProjection;
		private final PrecisionMeasure precisionMeasure;

		private TracesProcessor(ProjectedTrace processProjection, PrecisionMeasure precisionMeasure) {
			this.processProjection = processProjection;
			this.precisionMeasure = precisionMeasure;
		}

		public Void call() throws Exception {
			precisionMeasure.processTrace(processProjection);
			return null;
		}
	}

	private final PrecisionConfig config;

	private final StateFactory stateFactory;
	private final ProcessProjector processProjector;
	private final XLog log;
	private final DataPetriNet net;

	public DataAwarePrecision(PrecisionConfig config, DataPetriNet net, XLog log, XEventClasses eventClasses)
			throws ProcessProjectionException {
		this(config, net, log, new ProcessProjectorFittingImpl(net, log, eventClasses, config.getExtraAttributes(),
				config.getExecutors()), new StateFactoryDPNImpl());
	}

	public DataAwarePrecision(PrecisionConfig config, DataPetriNet net, XLog log, ProcessProjector projector) {
		this(config, net, log, projector, new StateFactoryDPNImpl());
	}

	public DataAwarePrecision(PrecisionConfig config, DataPetriNet net, XLog log, ProcessProjector projector,
			StateFactory stateFactory) {
		super();
		this.config = config;
		this.net = net;
		this.log = log;
		this.processProjector = projector;
		this.stateFactory = stateFactory;
	}

	public PrecisionResult computeDataAwarePrecision(ResultMode resultMode) throws PrecisionMeasureException {

		PrecisionMeasure precisionMeasure;
		if (config.getPrecisionMeasure() == PrecisionType.RESOURCE) {
			precisionMeasure = new ResourcePrecisionMeasureImpl(config, processProjector, stateFactory, net);
		} else {
			precisionMeasure = new ActivityPrecisionMeasureImpl(config, processProjector, stateFactory, net);
		}

		int concurrentThreads = config.getConcurrentThreads();
		int numTraces = log.size();

		ExecutorService executor = Executors.newFixedThreadPool(concurrentThreads);

		List<Future<Void>> results = new ArrayList<>(numTraces);

		for (Iterator<ProjectedTrace> projectedTraces = processProjector.projectedTraces(); projectedTraces.hasNext();) {

			results.add(executor.submit(new TracesProcessor(projectedTraces.next(), precisionMeasure)));

		}

		executor.shutdown();

		try {
			executor.awaitTermination(10, TimeUnit.MINUTES);
		} catch (InterruptedException e) {
			throw new PrecisionMeasureException(e);
		}

		for (Future<Void> f : results) {
			try {
				f.get();
			} catch (InterruptedException | ExecutionException e) {
				throw new PrecisionMeasureException(e);
			}
		}

		precisionMeasure.collectResults(resultMode);

		return new PrecisionResultImpl(precisionMeasure);

	}
}