package org.processmining.dataawarereplayer.precision.visualizer;

import java.awt.Color;
import java.awt.Dimension;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingWorker;
import javax.swing.table.DefaultTableModel;

import org.processmining.dataawarereplayer.precision.PrecisionResult;
import org.processmining.dataawarereplayer.precision.state.PrecisionState;
import org.processmining.dataawarereplayer.precision.state.StateTransition;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.utils.MarkingsHelper;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DPNAsDot;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DecorationKey;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.GuardDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.PlaceDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.VariableDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.RatioAwareDotPanel;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTableWithoutPanel;
import org.processmining.framework.util.ui.widgets.ProMTextArea;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.visualisation.listeners.SelectionChangedListener;

import com.google.common.collect.Lists;
import com.google.common.collect.SetMultimap;

final class PrecisionResultVisualizerPanel extends ProMSplitPane {

	private static final long serialVersionUID = 8834504180709309121L;

	private static final NumberFormat NF_FLOAT = NumberFormat.getNumberInstance();
	private static final NumberFormat NF_INT = NumberFormat.getIntegerInstance();

	private static final String PLACE_PRECISION_MESSAGE = "Precision {0} ({1} / {2})";
	private static final String STANDARD_MESSAGE = "Click on a transition/place to get more information.";

	private final DataPetriNet net;
	private final PrecisionResult precisionResult;

	private final RatioAwareDotPanel panel;
	private final ProMSplitPane topSplitpane;
	private final JTable placeTable;
	private final DefaultTableModel placeTableModel;

	private Collection<PrecisionTrace> overallTraces;

	private ProMTraceList<PrecisionTrace> traceList;

	public PrecisionResultVisualizerPanel(PluginContext context, DataPetriNet net, PrecisionResult precisionResult) {
		super(ProMSplitPane.VERTICAL_SPLIT);
		this.net = net;
		this.precisionResult = precisionResult;

		setResizeWeight(1.0);
		setOneTouchExpandable(true);

		Map<AbstractGraphElement, Map<DecorationKey, Object>> viewSpecific = new HashMap<>();

		for (Place p : net.getPlaces()) {
			double precision = precisionResult.getPrecision(p);
			Map<DecorationKey, Object> deco = new HashMap<DPNGraphvizConverter.DecorationKey, Object>();
			deco.put(DecorationKey.FILLCOLOR, colorGradient(Color.RED, Color.ORANGE, (float) precision));
			viewSpecific.put(p, deco);
		}

		final DPNAsDot convertedDPN = DPNGraphvizConverter.convertDPN(net,
				MarkingsHelper.getInitialMarkingOrEmpty(context, net),
				MarkingsHelper.getFinalMarkingsOrEmpty(context, net), GuardDisplayMode.HIDDEN,
				VariableDisplayMode.AUTO_LAYOUT, PlaceDisplayMode.DETAILED, GraphDirection.leftRight, viewSpecific);
		Dot dot = convertedDPN.getDot();

		panel = new RatioAwareDotPanel(dot);
		panel.setToolTipText(STANDARD_MESSAGE);
		panel.addSelectionChangedListener(new SelectionChangedListener<DotElement>() {

			public void selectionChanged(Set<DotElement> selectedElements) {
				Collection<DirectedGraphNode> selectedNodes = convertElements(selectedElements);
				if (!selectedNodes.isEmpty()) {
					final DirectedGraphNode node = selectedNodes.iterator().next();
					if (node instanceof Transition) {
						setInfoArea(createTransitionDetails((Transition) node));
					} else if (node instanceof Place) {
						SwingWorker<Collection<PrecisionTrace>, Void> precisionTraceWorker = new SwingWorker<Collection<PrecisionTrace>, Void>() {

							protected Collection<PrecisionTrace> doInBackground() throws Exception {
								return getPrecisionTraces((Place) node);
							}

							protected void done() {
								traceList.clear();
								try {
									traceList.addAll(get());
								} catch (InterruptedException | ExecutionException e) {
									ProMUIHelper.showErrorMessage(e.getMessage(), "Error creating traces");
								}
								traceList.sort(new Comparator<PrecisionTrace>() {

									public int compare(PrecisionTrace o1, PrecisionTrace o2) {
										return Long.compare(o2.getFrequency(), o1.getFrequency());
									}
								});
								setInfoArea(traceList);
							}

						};
						precisionTraceWorker.execute();
					} else {
						traceList.clear();
						traceList.addAll(overallTraces);
						traceList.sort(new Comparator<PrecisionTrace>() {

							public int compare(PrecisionTrace o1, PrecisionTrace o2) {
								return Long.compare(o2.getFrequency(), o1.getFrequency());
							}
						});
						setInfoArea(traceList);
					}
				} else {
					traceList.clear();
					traceList.addAll(overallTraces);
					traceList.sort(new Comparator<PrecisionTrace>() {

						public int compare(PrecisionTrace o1, PrecisionTrace o2) {
							return Long.compare(o2.getFrequency(), o1.getFrequency());
						}
					});
					setInfoArea(traceList);
				}
			}

			private Collection<DirectedGraphNode> convertElements(Set<DotElement> selectedElements) {
				List<DirectedGraphNode> convertedElements = new ArrayList<DirectedGraphNode>();
				for (DotElement element : selectedElements) {
					convertedElements.add(convertedDPN.getMapping().inverse().get(element));
				}
				return convertedElements;
			}

		});
		panel.setPreferredSize(new Dimension(Short.MAX_VALUE, 400));

		topSplitpane = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		topSplitpane.setResizeWeight(1.0);
		topSplitpane.setOneTouchExpandable(true);
		topSplitpane.setPreferredSize(new Dimension(Short.MAX_VALUE, 300));

		placeTableModel = new DefaultTableModel(new String[] { "Place", "Precision", "Observed", "Possible" }, 0);

		final List<Place> sortedPlaces = sortPlaces(net);
		for (Place p : sortedPlaces) {
			double precision = precisionResult.getPrecision(p);
			placeTableModel.addRow(new String[] { p.getLabel(), NumberFormat.getNumberInstance().format(precision),
					String.valueOf(precisionResult.getObservedContinuations(p)),
					String.valueOf(precisionResult.getPossibleContinuations(p)) });
		}
		placeTableModel.addRow(new String[] { "Model Precision",
				NumberFormat.getNumberInstance().format(precisionResult.getPrecision()) });

		placeTable = new ProMTableWithoutPanel(placeTableModel);
		setupTableSize();
		placeTable.setBackground(Color.WHITE);
		placeTable.setMinimumSize(new Dimension(600, Short.MAX_VALUE));

		overallTraces = getPrecisionTraces();
		traceList = PrecisionStateListViewUtils.createPrecisionTraceList(overallTraces);

		topSplitpane.setLeftComponent(panel);
		JScrollPane tableScrollPane = new JScrollPane(placeTable);
		tableScrollPane.setMinimumSize(new Dimension(600, Short.MAX_VALUE));
		topSplitpane.setRightComponent(tableScrollPane);

		setTopComponent(topSplitpane);
		setBottomComponent(traceList);
	}

	private void setupTableSize() {
		placeTable.getColumnModel().getColumn(0).setPreferredWidth(150);
		placeTable.getColumnModel().getColumn(1).setPreferredWidth(50);
		placeTable.getColumnModel().getColumn(2).setPreferredWidth(50);
		placeTable.getColumnModel().getColumn(3).setPreferredWidth(50);
	}

	private final void setInfoArea(JComponent component) {
		setBottomComponent(component);
	}

	private final ProMTextArea createTransitionDetails(Transition transition) {
		ProMTextArea text = new ProMTextArea(false);
		text.setText(STANDARD_MESSAGE);
		text.setLineWrap(true);
		if (transition instanceof PNWDTransition && ((PNWDTransition) transition).getGuardExpression() != null) {
			text.setText("Transition: " + transition.getLabel() + "\nGuard: "
					+ ((PNWDTransition) transition).getGuardExpression().toPrettyString(1));
		} else {
			text.setText("Transition: " + transition.getLabel() + "\nGuard: NO Guard");
		}
		return text;
	}

	private final Collection<PrecisionTrace> getPrecisionTraces(final Place node) {
		SetMultimap<PrecisionState, StateTransition> possibleStateTransitions = precisionResult
				.getPossibleStateTransitions(node);
		SetMultimap<PrecisionState, StateTransition> observedStateTransitions = precisionResult
				.getObservedStateTransitions(node);

		Collection<PrecisionTrace> precisionTraces = PrecisionStateListViewUtils.createPrecisionTraces(precisionResult,
				possibleStateTransitions, observedStateTransitions);
		return precisionTraces;
	}

	private final Collection<PrecisionTrace> getPrecisionTraces() {
		SetMultimap<PrecisionState, StateTransition> possibleStateTransitions = precisionResult
				.getPossibleStateTransitions();
		SetMultimap<PrecisionState, StateTransition> observedStateTransitions = precisionResult
				.getObservedStateTransitions();

		Collection<PrecisionTrace> precisionTraces = PrecisionStateListViewUtils.createPrecisionTraces(precisionResult,
				possibleStateTransitions, observedStateTransitions);
		return precisionTraces;
	}

	@SuppressWarnings("unused")
	private static Place getSelectedPlace(Collection<DirectedGraphNode> selectNodes) {
		Iterator<DirectedGraphNode> iterator = selectNodes.iterator();
		Place node = null;
		while (iterator.hasNext() && node == null) {
			DirectedGraphNode nextNode = iterator.next();
			if (nextNode instanceof Place) {
				return (Place) nextNode;
			}
		}
		return null;
	}

	private static Color colorGradient(Color darkColor, Color lightColor, float factor) {
		float[] hsb = Color.RGBtoHSB(darkColor.getRed(), darkColor.getGreen(), darkColor.getBlue(), null);
		float[] hsb2 = Color.RGBtoHSB(lightColor.getRed(), lightColor.getGreen(), lightColor.getBlue(), null);
		float brightness = 1f;
		float saturation = 1 - factor;
		float minHueFactor = 0.0f;
		float maxHueFactor = 1.0f;
		float hue = hsb[0] + (hsb2[0] - hsb[0]) * ((factor - minHueFactor) / (maxHueFactor - minHueFactor));
		return Color.getHSBColor(hue, saturation, brightness);
	}

	private List<Place> sortPlaces(PetrinetGraph net) {
		final ArrayList<Place> places = Lists.newArrayList(net.getPlaces());
		Collections.sort(places, new Comparator<Place>() {

			public int compare(Place o1, Place o2) {
				return o1.getLabel() != null && o2.getLabel() != null ? o1.getLabel().compareTo(o2.getLabel())
						: o1.compareTo(o2);
			}
		});
		return places;
	}

}