package org.processmining.plugins.DataConformance.Analyze.PetriNet;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XTrace;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.models.FunctionEstimator.Type;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.DataAlignment.PetriNet.ResultReplayPetriNetWithData;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

public class DecisionTreeCreator {

	private static final String MOVEINLOG = "Move in Log";
	private static final String MOVEINMODEL = "Move in Process";
	private static final String MOVEINBOTHSAMEASSIGNMENT = "Move in Both without incorrect write operations";
	private static final String MOVEINBOTHDIFFERENTASSIGNMENT = "Move in Both with incorrect write operations";
	public static final String classAttribute = "TYPE OF MOVE";

	/**
	 * Creates a {@link TupleSet} for the specified {@link PetrinetNode} only
	 * 
	 * @param node
	 * @param net
	 * @param logReplayResult
	 * @param activityMapping
	 * @return
	 */
	public static TupleSet createInstanceSet(PetrinetNode node, DataPetriNet net,
			ResultReplayPetriNetWithData logReplayResult, TransEvClassMapping activityMapping) {

		Map<String, Type> attributeNames = new HashMap<>();
		for (DataElement dataElem : net.getVariables()) {
			Type t = null;
			Class classType = dataElem.getType();
			if (classType == String.class)
				t = Type.LITERAL;
			else if (classType == Integer.class || classType == Long.class)
				t = Type.DISCRETE;
			else if (classType == Boolean.class)
				t = Type.BOOLEAN;
			else if (classType == Float.class || classType == Double.class)
				t = Type.CONTINUOS;
			else if (classType == Date.class)
				t = Type.TIMESTAMP;
			else
				assert (false);
			attributeNames.put(dataElem.getVarName(), t);
			attributeNames.put(dataElem.getVarName().concat("'"), t);
		}
		for (Transition t : net.getTransitions()) {
			attributeNames.put("#".concat(t.getLabel()), Type.DISCRETE);
		}

		Map<String, Set<String>> literalAttributeValues = new HashMap<>();
		Set<String> classValue = new HashSet<>();
		classValue.add(MOVEINLOG);
		classValue.add(MOVEINMODEL);
		classValue.add(MOVEINBOTHSAMEASSIGNMENT);
		classValue.add(MOVEINBOTHDIFFERENTASSIGNMENT);
		literalAttributeValues.put(classAttribute, classValue);

		TupleSet tupleSet = new TupleSet(attributeNames, classAttribute, literalAttributeValues);
		XConceptExtension conceptExt = XConceptExtension.instance();
		
		for (XTrace trace : logReplayResult.getAlignedLog()) {
		
			Alignment alignment = logReplayResult.getAlignmentByTraceName(conceptExt.extractName(trace));
			
			if (alignment != null) {
			
				HashMap<String, Object> varValue = new HashMap<>();
				
				Iterator<ExecutionStep> logIter = alignment.getLogTrace().iterator();
				Iterator<ExecutionStep> processIter = alignment.getProcessTrace().iterator();
				Iterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().iterator();
				
				HashMap<String, Integer> activityNumExecutions = new HashMap<>();
				
				for (Transition t : net.getTransitions()) {
					activityNumExecutions.put(t.getLabel(), 0);
				}
				
				// Add trace attributes
				for (Entry<String, XAttribute> attribute : trace.getAttributes().entrySet()) {
					Object value = getValue(attribute.getValue(), literalAttributeValues, attributeNames);
					varValue.put(attribute.getKey(), value);
				}
				
				while (logIter.hasNext()) {
					
					ExecutionStep logStep = logIter.next();
					ExecutionStep procStep = processIter.next();
					String activityName = null;

					if (!procStep.equals(ExecutionStep.bottomStep)) {
						activityName = procStep.getActivity();
					} else {
						activityName = discoverProcessActivity(logStep.getActivity(), activityMapping);
					}
					
					if (activityName == null) {
						continue;
					}
					
					DataAwareStepTypes stepType = stepIter.next();
					
					Transition t = findTransition(net, activityName);
					
					if (t == null || t.isInvisible()) {
						continue;
					}
						
					
					if (t.equals(node)) {
				
						HashMap<String, Object> instance = (HashMap<String, Object>) varValue.clone();
						
						for (Entry<String, Integer> activityEntry : activityNumExecutions.entrySet()) {
							instance.put("#".concat(activityEntry.getKey()), activityEntry.getValue());
						}
							
						switch (stepType) {
							case L:
								instance.put(classAttribute, MOVEINLOG);
								break;
							case LMGOOD :
								instance.put(classAttribute, MOVEINBOTHSAMEASSIGNMENT);
								break;
							case LMNOGOOD :
								instance.put(classAttribute, MOVEINBOTHDIFFERENTASSIGNMENT);
								break;
							case MINVI :
							case MREAL :
								instance.put(classAttribute, MOVEINMODEL);
								break;
						}
						
						for (Entry<String, Object> entry : logStep.entrySet()) {
							instance.put(entry.getKey().concat("'"), entry.getValue());
						}
						
						tupleSet.addInstance(instance);
						
					}
					
					Integer numExecutions = activityNumExecutions.get(activityName);
					activityNumExecutions.put(activityName, ++numExecutions);
					for (Entry<String, Object> entry : logStep.entrySet()) {
						if (attributeNames.get(entry.getKey()) == Type.LITERAL) {
							Set<String> list = literalAttributeValues.get(entry.getKey());
							if (list == null) {
								list = new HashSet<>();
								literalAttributeValues.put(entry.getKey(), list);
								literalAttributeValues.put(entry.getKey().concat("'"), list);
							}
							list.add((String) entry.getValue());
						}
					}
					
					varValue.putAll(logStep);
				}
			}
		}
		
		return tupleSet;
		
	}
	
	/**
	 * Creates {@link TupleSet}s for all {@link Transition} of the Net.
	 * 
	 * @param net
	 * @param logReplayResult
	 * @param activityMapping
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static Map<PetrinetNode, TupleSet> createInstances(DataPetriNet net,
			ResultReplayPetriNetWithData logReplayResult, TransEvClassMapping activityMapping) {
		
		Map<String, Type> attributeNames = new HashMap<>();
		for (DataElement dataElem : net.getVariables()) {
			Type t = null;
			Class classType = dataElem.getType();
			if (classType == String.class)
				t = Type.LITERAL;
			else if (classType == Integer.class || classType == Long.class)
				t = Type.DISCRETE;
			else if (classType == Boolean.class)
				t = Type.BOOLEAN;
			else if (classType == Float.class || classType == Double.class)
				t = Type.CONTINUOS;
			else if (classType == Date.class)
				t = Type.TIMESTAMP;
			else
				assert (false);
			attributeNames.put(dataElem.getVarName(), t);
			attributeNames.put(dataElem.getVarName().concat("'"), t);
		}
		for (Transition t : net.getTransitions())
			attributeNames.put("#".concat(t.getLabel()), Type.DISCRETE);

		Map<String, Set<String>> literalAttributeValues = new HashMap<>();
		Set<String> classValue = new HashSet<>();
		classValue.add(MOVEINLOG);
		classValue.add(MOVEINMODEL);
		classValue.add(MOVEINBOTHSAMEASSIGNMENT);
		classValue.add(MOVEINBOTHDIFFERENTASSIGNMENT);
		literalAttributeValues.put(classAttribute, classValue);

		HashMap<PetrinetNode, TupleSet> retValue = new HashMap<>();
		for (Transition node : net.getTransitions()) {
			if (!node.isInvisible())
				retValue.put(node, new TupleSet(attributeNames, classAttribute, literalAttributeValues));
		}
		XConceptExtension conceptExt = XConceptExtension.instance();
		for (XTrace trace : logReplayResult.getAlignedLog()) {
			Alignment alignment = logReplayResult.getAlignmentByTraceName(conceptExt.extractName(trace));
			if (alignment != null) {
				HashMap<String, Object> varValue = new HashMap<>();
				Iterator<ExecutionStep> logIter = alignment.getLogTrace().iterator();
				Iterator<ExecutionStep> processIter = alignment.getProcessTrace().iterator();
				Iterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().iterator();
				HashMap<String, Integer> activityNumExecutions = new HashMap<>();
				for (Transition t : net.getTransitions())
					activityNumExecutions.put(t.getLabel(), 0);
				for (Entry<String, XAttribute> attribute : trace.getAttributes().entrySet()) {
					Object value = getValue(attribute.getValue(), literalAttributeValues, attributeNames);
					varValue.put(attribute.getKey(), value);
				}
				while (logIter.hasNext()) {
					ExecutionStep logStep = logIter.next();
					ExecutionStep procStep = processIter.next();
					String activityName = null;
					if (!procStep.equals(ExecutionStep.bottomStep))
						activityName = procStep.getActivity();
					else
						activityName = discoverProcessActivity(logStep.getActivity(), activityMapping);
					if (activityName == null)
						continue;
					DataAwareStepTypes stepType = stepIter.next();
					Transition t = findTransition(net, activityName);
					if (t == null || t.isInvisible())
						continue;
					HashMap<String, Object> instance = (HashMap<String, Object>) varValue.clone();
					for (Entry<String, Integer> activityEntry : activityNumExecutions.entrySet())
						instance.put("#".concat(activityEntry.getKey()), activityEntry.getValue());
					switch (stepType) {
						case L :
							instance.put(classAttribute, MOVEINLOG);
							break;
						case LMGOOD :
							instance.put(classAttribute, MOVEINBOTHSAMEASSIGNMENT);
							break;
						case LMNOGOOD :
							instance.put(classAttribute, MOVEINBOTHDIFFERENTASSIGNMENT);
							break;
						case MINVI :
						case MREAL :
							instance.put(classAttribute, MOVEINMODEL);
							break;
					}
					for (Entry<String, Object> entry : logStep.entrySet()) {
						instance.put(entry.getKey().concat("'"), entry.getValue());
					}
					retValue.get(t).addInstance(instance);
					Integer numExecutions = activityNumExecutions.get(activityName);
					activityNumExecutions.put(activityName, ++numExecutions);
					for (Entry<String, Object> entry : logStep.entrySet()) {
						if (attributeNames.get(entry.getKey()) == Type.LITERAL) {
							Set<String> list = literalAttributeValues.get(entry.getKey());
							if (list == null) {
								list = new HashSet<>();
								literalAttributeValues.put(entry.getKey(), list);
								literalAttributeValues.put(entry.getKey().concat("'"), list);
							}
							list.add((String) entry.getValue());
						}
					}
					varValue.putAll(logStep);
				}
			}
		}
		return retValue;
	}

	private static String discoverProcessActivity(String logActivityName, TransEvClassMapping activityMapping) {
		for (Entry<Transition, XEventClass> entry : activityMapping.entrySet()) {
			String id = entry.getValue().getId();
			if (id.lastIndexOf("+") > 0)
				id = id.substring(0, id.lastIndexOf("+"));

			if (id.equals(logActivityName))
				return entry.getKey().getLabel();
		}
		return null;
	}

	private static Transition findTransition(PetrinetGraph net, String activity) {
		for (Transition trans : net.getTransitions()) {
			if (trans.getLabel().equals(activity))
				return trans;
		}
		return null;
	}

	private static Object getValue(XAttribute xAttribute, Map<String, Set<String>> literalAttributeValues,
			Map<String, Type> attributeNames) {
		if (xAttribute instanceof XAttributeDiscrete)
			return ((XAttributeDiscrete) xAttribute).getValue();
		if (xAttribute instanceof XAttributeContinuous)
			return ((XAttributeContinuous) xAttribute).getValue();
		if (xAttribute instanceof XAttributeTimestamp)
			return ((XAttributeTimestamp) xAttribute).getValue();
		if (xAttribute instanceof XAttributeLiteral) {
			String value = ((XAttributeLiteral) xAttribute).getValue();
			if (attributeNames.get(xAttribute.getKey()) == Type.BOOLEAN)
				return Boolean.parseBoolean(value);
			Set<String> list = literalAttributeValues.get(xAttribute.getKey());
			if (list == null) {
				list = new HashSet<>();
				literalAttributeValues.put(xAttribute.getKey(), list);
			}
			list.add(value);
			return value;
		}
		throw new RuntimeException("Could not get value from " + xAttribute.toString());
	}

}
