package org.processmining.plugins.DataConformance.framework;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.balancedconformance.dataflow.MILPProblemBuilder;

import com.google.common.collect.Ordering;
import com.google.common.math.DoubleMath;

// TODO FM, turn this into an interface and allow for different implementations
/**
 * A move in either the process or log projection of an {@link Alignment}.
 *
 */
public class ExecutionStep extends HashMap<String, Object> {

	public static class ExecutionStepBuilder {

		private ExecutionStep executionStep;

		public ExecutionStepBuilder activity(String activity) {
			executionStep = new ExecutionStep(activity);
			return this;
		}

		public ExecutionStepBuilder bottomStep() {
			executionStep = bottomStep;
			return this;
		}

		public ExecutionStepBuilder variable(String key, Object value) {
			if (executionStep == null) {
				throw new IllegalStateException("Should use 'activity' beforehand!");
			}
			if (executionStep == bottomStep) {
				throw new IllegalStateException("Cannot assign variable values to bottom step!");
			}
			executionStep.put(key, value);
			return this;
		}
		
		public ExecutionStepBuilder invisible() {
			executionStep.setInvisible(true);
			return this;
		}

		public ExecutionStep build() {
			return executionStep;
		}

	}

	public static ExecutionStepBuilder builder() {
		return new ExecutionStepBuilder();
	}

	// We need a bit more precision than the epsilon used in the ILP problems
	private static final double COMPARISON_PRECISION = MILPProblemBuilder.RESULT_EPSILON  / 10;

	private static final long serialVersionUID = -8838709839593170889L;

	public static final Ordering<Map.Entry<String, Object>> DEFAULT_VARIABLE_ORDERING = new Ordering<Map.Entry<String, Object>>() {

		public int compare(Map.Entry<String, Object> arg0, Map.Entry<String, Object> arg1) {
			return arg0.getKey().compareTo(arg1.getKey());
		}
	};

	private final String activity;
	private final Object activityObject;
	private boolean invisible = false;

	public static ExecutionStep bottomStep = new ExecutionStep() {
		
		private static final long serialVersionUID = 1558682239472411125L;

		/* (non-Javadoc)
		 * @see java.util.AbstractMap#toString()
		 */
		public String toString() {
			return "BOTTOM-STEP";
		}

		public Object put(String key, Object value) {
			throw new UnsupportedOperationException("Immutable Step");
		}

		public void putAll(Map<? extends String, ? extends Object> m) {
			throw new UnsupportedOperationException("Immutable Step");
		}
		
	};

	public ExecutionStep(String activity, Object activityObject) {
		if (activity == null)
			throw new IllegalArgumentException("The activity name should not be NULL");
		this.activity = activity;
		this.activityObject = activityObject;
	}

	public ExecutionStep(String activity) {
		if (activity == null)
			throw new IllegalArgumentException("The activity name should not be NULL");
		this.activity = activity;
		activityObject = null;
	}

	public ExecutionStep() {
		activity = null;
		activityObject = null;
	}

	/**
	 * @return the activity or event label
	 */
	public String getActivity() {
		return activity;
	}

	/**
	 * @return if available, the underlying Transition or XEvent (may be NULL)
	 */
	public Object getActivityObject() {
		return activityObject;
	}

	/* (non-Javadoc)
	 * @see java.util.AbstractMap#toString()
	 */
	public String toString() {
		if (activity != null) {
			return activity + " " + super.toString();
		} else {
			return super.toString();
		}
	}

	/* (non-Javadoc)
	 * @see java.util.AbstractMap#hashCode()
	 */
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((activity == null) ? 0 : activity.hashCode());
		result = prime * result + (invisible ? 1231 : 1237);
		return result;
	}

	/* (non-Javadoc)
	 * @see java.util.AbstractMap#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (getClass() != obj.getClass())
			return false;
		ExecutionStep other = (ExecutionStep) obj;
		if (activity == null) {
			if (other.activity != null)
				return false;
		} else if (!activity.equals(other.activity))
			return false;
		if (invisible != other.invisible)
			return false;
		if (!equalsAssignmentsOnly(other))
			return false;
		return true;
	}

	public boolean equalsAssignmentsOnly(ExecutionStep other) {
		if (size() != other.size()) {
			return false;
		}
		Iterator<Map.Entry<String, Object>> iter = entrySet().iterator();
		while (iter.hasNext()) {
			Map.Entry<String, Object> e = iter.next();
			String varName = e.getKey();
			Object varValue = e.getValue();
			Object otherValue = other.get(varName);
			if (otherValue == null) {
				// Not Available
				return false;
			}
			if (varValue instanceof Double || varValue instanceof Float) {
				if (!DoubleMath.fuzzyEquals(((Number) varValue).doubleValue(), ((Number) otherValue).doubleValue(),
						COMPARISON_PRECISION)) {
					return false;
				}
			} else {
				if (varValue == null || otherValue == null || !varValue.equals(otherValue)) {
					return false;
				}
			}
		}
		return true;
	}

	public void setInvisible(boolean invisible) {
		this.invisible = invisible;
	}

	public boolean isInvisible() {
		return invisible;
	}

}
