/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.cache;

import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClasses;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;

import com.google.common.collect.Maps;

import nl.tue.astar.Tail;

/**
 * Assigns guards and write operations of transitions unique numbers. 
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataViewBuilder {

	private final Map<GuardExpression, Short> duplicateGuardMap;
	private final Map<Set<String>, Short> duplicateWriteOperationsMap;
	
	private final short[] transitionMap;
	private final short[] guardMap;
	private short maxIndex;
	
	private final short numEventClasses;
	private boolean hasWriteOps = false;

	public DataViewBuilder(DataPetriNet net, AbstractPDelegate<? extends Tail> delegate, XEventClasses classes, LogMapping logMapping) {
		if (classes.size() >= Short.MAX_VALUE) {
			throw new IllegalArgumentException("Cannot handle more than "+Short.MAX_VALUE+ " event classes!");
		}
		numEventClasses = (short) classes.size();
		duplicateGuardMap = Maps.newHashMapWithExpectedSize(net.getTransitions().size());
		duplicateWriteOperationsMap = Maps.newHashMapWithExpectedSize(net.getTransitions().size());
		
		transitionMap = new short[net.getTransitions().size()];
		guardMap = new short[net.getTransitions().size()];
		short count = 1;
		for (Transition t : net.getTransitions()) {
			Set<String> writeOperations = logMapping.getVariablesToWrite().get(t);
			if (!writeOperations.isEmpty()) {
				hasWriteOps = true;
				// Each transition that writes variables gets a unique number
				Short writeOpsIndex = duplicateWriteOperationsMap.get(writeOperations);
				if (writeOpsIndex == null) {
					writeOpsIndex = count++;
				}
				duplicateWriteOperationsMap.put(writeOperations, writeOpsIndex);
				transitionMap[delegate.getIndexOf(t)] = writeOpsIndex;	
			} else {
				// No write operation mapped to -1 as this Transition it is not important
				transitionMap[delegate.getIndexOf(t)] = -1;
			}
			GuardExpression guard = logMapping.getGuardExpression(t);
			if (guard != null) {
				Short guardIndex = duplicateGuardMap.get(guard);
				if (guardIndex == null) {
					guardIndex = count++;
				}
				duplicateGuardMap.put(guard, guardIndex);
				// Each unique guard in the net gets a unique number
				guardMap[delegate.getIndexOf(t)] = guardIndex;
			} else {
				guardMap[delegate.getIndexOf(t)] = -1;
			}
		}
		maxIndex = count;
	}

	public short buildWriteDataView(int transitionIndex) {
		return transitionMap[transitionIndex];
	}

	public short buildGuardDataView(int transitionIndex) {
		return guardMap[transitionIndex];
	}
	
	public short buildEventDataView(short writeDataView, int eventIndex) {
		int eventDataView = nextSafeIndex() 
				+ (writeDataView * numEventClasses) // assign every write operation its own address space, 
													// event variables might be mapped to different writes 
				+ eventIndex;
		if (eventDataView >= Short.MAX_VALUE) {
			throw new RuntimeException("Too many events, does not support traces longer than "+ (Short.MAX_VALUE - nextSafeIndex()));
		}
		return (short) eventDataView;
	}
	
	public short nextSafeIndex() {		
		return maxIndex;
	}

	public boolean hasWriteOps() {
		return hasWriteOps;
	}

}
