/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.config;

import java.util.Collections;
import java.util.Map;

import nl.tue.astar.AStarThread.Canceller;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserver;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverNoOpImpl;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

/**
 * Base class for DataConformancePlus configurations
 * 
 * @author F. Mannhardt
 * 
 */
public class DataConformancePlusConfiguration {

	//*************** Mandatory fields *****************

	/**
	 * Mapping of Transition to Event Class
	 */
	private TransEvClassMapping activityMapping = null;

	private Map<XEventClass, Integer> mapEvClass2Cost = null;

	private Map<Transition, Integer> mapTrans2Cost = null;

	private Marking initialMarking = null;

	private Marking[] finalMarkings = null;

	/**
	 * Mapping from variable name in DPN-net to variable name in event log
	 */
	private Map<String, String> variableMapping = null;

	/**
	 * Mapping from variable name in DPN-net to cost
	 */
	private VariableMatchCosts variableCost = null;

	//*************** Optional fields *****************

	/**
	 * Observer of the DataConformance checking
	 */
	private DataConformancePlusObserver observer = new DataConformancePlusObserverNoOpImpl();

	/**
	 * Canceller for the PNReplayer
	 */
	private Canceller controlFlowAlignmentCanceller = new Canceller() {

		public boolean isCancelled() {
			return false;
		}
	};

	/**
	 * Mapping from variable name in DPN-net to lower bound
	 */
	private Map<String, Object> lowerBounds = Collections.emptyMap();

	/**
	 * Mapping from variable name in DPN-net to upper bound
	 */
	private Map<String, Object> upperBounds = Collections.emptyMap();

	/**
	 * Mapping from variable name in DPN-net to a default value that is assumed
	 * at the initial state. In case there is not default value, the variable is
	 * assumed to be 'free'.
	 */
	private Map<String, Object> defaultValues = Collections.emptyMap();

	/**
	 * Number of concurrent threads to spawn
	 */
	private int concurrentThreads = Runtime.getRuntime().availableProcessors();

	/**
	 * Use a built-in cache that remembers the result of the recently calculated
	 * ILP problems
	 */
	private boolean activateDataViewCache = true;

	/**
	 * Clear/Keep the search space after each trace
	 */
	private boolean keepControlFlowSearchSpace = true;

	private boolean keepDataFlowSearchSpace = true;

	private boolean usePartialOrder = false;

	/**
	 * Creates a new configuration with default settings, you still need to set
	 * the mandatory parameters (see
	 * {@link #DataConformancePlusConfiguration(Marking, Marking[], TransEvClassMapping, Map, Map, Map, VariableMatchCosts, Map)}
	 * manually.
	 */
	public DataConformancePlusConfiguration() {
		super();
	}

	/**
	 * Creates a new configuration with the supplied mandatory parameters. All
	 * other configuration options are set to reasonable defaults.
	 * 
	 * @param initialMarking
	 * @param finalMarkings
	 * @param activityMapping
	 *            Mapping of Transition to Event Class
	 * @param mapEvClass2Cost
	 * @param mapTrans2Cost
	 * @param variableMapping
	 *            Mapping from variable name in DPN-net to variable name in
	 *            event log
	 * @param variableCost
	 *            Mapping from variable name in DPN-net to cost
	 * @param varType
	 *            Mapping from variable name in DPN-net to Java class
	 */
	public DataConformancePlusConfiguration(Marking initialMarking, Marking[] finalMarkings,
			TransEvClassMapping activityMapping, Map<XEventClass, Integer> mapEvClass2Cost,
			Map<Transition, Integer> mapTrans2Cost, Map<String, String> variableMapping,
			VariableMatchCosts variableCost, boolean usePartialOrder) {
		super();
		this.initialMarking = initialMarking;
		this.finalMarkings = finalMarkings;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.mapTrans2Cost = mapTrans2Cost;
		this.activityMapping = activityMapping;
		this.variableMapping = variableMapping;
		this.variableCost = variableCost;
		this.usePartialOrder = usePartialOrder;
	}

	/**
	 * Copy constructor that takes its parameters from the supplied
	 * configuration.
	 * 
	 * @param config
	 *            to copy (shallow-clone) from
	 */
	public DataConformancePlusConfiguration(DataConformancePlusConfiguration config) {
		super();
		this.initialMarking = config.getInitialMarking();
		this.finalMarkings = config.getFinalMarkings();
		this.mapEvClass2Cost = config.getMapEvClass2Cost();
		this.mapTrans2Cost = config.getMapTrans2Cost();
		this.variableMapping = config.getVariableMapping();
		this.variableCost = config.getVariableCost();
		this.controlFlowAlignmentCanceller = config.getControlFlowAlignmentCanceller();
		this.observer = config.getObserver();
		this.lowerBounds = config.getLowerBounds();
		this.upperBounds = config.getUpperBounds();
		this.defaultValues = config.getDefaultValues();
		this.activityMapping = config.getActivityMapping();
		this.concurrentThreads = config.getConcurrentThreads();
		this.activateDataViewCache = config.isActivateDataViewCache();
		this.keepControlFlowSearchSpace = config.isKeepControlFlowSearchSpace();
		this.keepDataFlowSearchSpace = config.isKeepDataFlowSearchSpace();
		this.usePartialOrder = config.isUsePartialOrder();
	}

	public boolean isUsePartialOrder() {
		return usePartialOrder;
	}

	/**
	 * @return Map from variable name in DPN-net to variable name in event log
	 */
	public Map<String, String> getVariableMapping() {
		return variableMapping;
	}

	/**
	 * @param variableMapping
	 *            mapping from variable name in DPN-net to variable name in
	 *            event log
	 */
	public void setVariableMapping(Map<String, String> variableMapping) {
		this.variableMapping = variableMapping;
	}

	public VariableMatchCosts getVariableCost() {
		return variableCost;
	}

	/**
	 * @param variableCost
	 *            mapping from variable name in DPN-net to cost
	 */
	public void setVariableCost(VariableMatchCosts variableCost) {
		this.variableCost = variableCost;
	}

	public DataConformancePlusObserver getObserver() {
		return observer;
	}

	public void setObserver(DataConformancePlusObserver observer) {
		this.observer = observer;
	}

	public void setControlFlowAlignmentCanceller(Canceller canceller) {
		this.controlFlowAlignmentCanceller = canceller;
	}

	public Canceller getControlFlowAlignmentCanceller() {
		return controlFlowAlignmentCanceller;
	}

	/**
	 * @param lowerBounds
	 *            mapping from variable name in DPN-net to lower bound (helps
	 *            speeding up ILP solving)
	 */
	public void setLowerBounds(Map<String, Object> lowerBounds) {
		this.lowerBounds = lowerBounds;
	}

	/**
	 * @param upperBounds
	 *            mapping from variable name in DPN-net to upper bound (helps
	 *            speeding up ILP solving)
	 */
	public void setUpperBounds(Map<String, Object> upperBounds) {
		this.upperBounds = upperBounds;
	}

	public TransEvClassMapping getActivityMapping() {
		return activityMapping;
	}

	/**
	 * Configures the activity to event class mapping. Take care to use the same
	 * classifier here and for {@link #setMapEvClass2Cost(Map)}!
	 * 
	 * @param activityMapping
	 */
	public void setActivityMapping(TransEvClassMapping activityMapping) {
		this.activityMapping = activityMapping;
	}

	public Map<String, Object> getUpperBounds() {
		return upperBounds;
	}

	public Map<String, Object> getLowerBounds() {
		return lowerBounds;
	}

	public boolean isActivateDataViewCache() {
		return activateDataViewCache;
	}

	/**
	 * @param activateDataViewCache
	 *            use a built-in cache that remembers the result of the recently
	 *            calculated ILP problems
	 */
	public void setActivateDataViewCache(boolean activateDataViewCache) {
		this.activateDataViewCache = activateDataViewCache;
	}

	public int getConcurrentThreads() {
		return concurrentThreads;
	}

	/**
	 * @param concurrentThreads
	 *            number of concurrent threads to spawn
	 */
	public void setConcurrentThreads(int concurrentThreads) {
		this.concurrentThreads = concurrentThreads;
	}

	public boolean isKeepControlFlowSearchSpace() {
		return keepControlFlowSearchSpace;
	}

	/**
	 * @param keepSearchSpace
	 *            clear/Keep the search space after each trace
	 */
	public void setKeepControlFlowSearchSpace(boolean keepSearchSpace) {
		this.keepControlFlowSearchSpace = keepSearchSpace;
	}

	public boolean isKeepDataFlowSearchSpace() {
		return keepDataFlowSearchSpace;
	}

	/**
	 * @param keepSearchSpace
	 *            clear/Keep the search space after each trace
	 */
	public void setKeepDataFlowSearchSpace(boolean keepSearchSpace) {
		this.keepDataFlowSearchSpace = keepSearchSpace;
	}

	public Map<XEventClass, Integer> getMapEvClass2Cost() {
		return mapEvClass2Cost;
	}

	public void setMapEvClass2Cost(Map<XEventClass, Integer> mapEvClass2Cost) {
		this.mapEvClass2Cost = mapEvClass2Cost;
	}

	public Map<Transition, Integer> getMapTrans2Cost() {
		return mapTrans2Cost;
	}

	public void setMapTrans2Cost(Map<Transition, Integer> mapTrans2Cost) {
		this.mapTrans2Cost = mapTrans2Cost;
	}

	public Marking getInitialMarking() {
		return initialMarking;
	}

	public void setInitialMarking(Marking initialMarking) {
		this.initialMarking = initialMarking;
	}

	public Marking[] getFinalMarkings() {
		return finalMarkings;
	}

	public void setFinalMarkings(Marking[] finalMarkings) {
		this.finalMarkings = finalMarkings;
	}

	public Map<String, Object> getDefaultValues() {
		return defaultValues;
	}

	/**
	 * @param defaultValue
	 *            Mapping from variable name in DPN-net to a default value that
	 *            is assumed at the initial state. In case there is not default
	 *            value, the variable is assumed to be 'free'.
	 */
	public void setDefaultValues(Map<String, Object> defaultValue) {
		this.defaultValues = defaultValue;
	}

	public void setUsePartialOrders(boolean usePartialOrders) {
		this.usePartialOrder = usePartialOrders;
	}

}