/*
 * Adapted from PNetReplayer package
 */
package org.processmining.plugins.balancedconformance.controlflow.override;

import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.processmining.plugins.astar.petrinet.impl.PRecord;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.controlflow.adapter.AlignmentAdapter.PriorityQueueFactory;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.RandomizedQueue;
import org.processmining.plugins.balancedconformance.controlflow.override.storage.ConsideredStatesStoreImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;

import it.unimi.dsi.fastutil.Hash.Strategy;
import nl.tue.astar.AStarException;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.State;
import nl.tue.astar.impl.memefficient.StorageAwareDelegate;

/**
 * Returns the optimal control flow and data alignment by adding the data cost
 * of the current alignment prefix to the total cost of each record. This
 * implementation is NOT synchronized it should only be accesses by a single
 * thread at all times. It returns a {@link DRecord} instead of the normal
 * {@link PRecord} as we need to store some additional information.
 * 
 * @author F. Mannhardt
 * 
 * @param <H>
 *            expected to be a DHead
 * @param <T>
 */
public final class DataAwareGraphThread<H extends Head, T extends Tail> extends DataAwareThread<H, T> {

	private final boolean isRandomizedQueue;

	public DataAwareGraphThread(StorageAwareDelegate<H, T> delegate, H initialHead, StorageHandler<H, T> storageHandler,
			Trace trace, int maxStates, DataAlignmentAdapter adapter, LogMapping logMapping, TraceMapping traceMapping,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost,
			DataViewBuilder dataViewBuilder, boolean activateCache, boolean isUseImprovedMethod,
			ASynchronousMoveSorting sorting, PriorityQueueFactory queueFactory, DataStateStore dataStateStore, Strategy<DRecord> strategy) {
		super(delegate, initialHead, storageHandler, trace, maxStates, adapter, logMapping, traceMapping,
				mapEv2EventClass, mapEvClass2Cost, dataViewBuilder, activateCache, isUseImprovedMethod, sorting,
				queueFactory, dataStateStore);
		// Override field in superclass
		consideredStatesStore = new ConsideredStatesStoreImpl(strategy, 1024 * 8);
		isRandomizedQueue = queue instanceof RandomizedQueue;
	}

	@Override
	protected final void processMove(H head, T tail, Record rec, int modelMove, int movedEvent, int activity)
			throws AStarException {

		// First, construct the next head from the old head
		final H newHead = computeNextHead(rec, head, modelMove, movedEvent, activity);
		final long stateIndex;
		try {
			// Check whether the new head is already in search space
			stateIndex = storageHandler.getIndexOf(newHead);
		} catch (Exception e) {
			throw new AStarException(e);
		}

		// First, create control-flow record
		DRecord newRec = ((DRecord) rec).getNextRecord(delegate, trace, newHead, stateIndex, modelMove, movedEvent,
				activity);
		
		traversedArcCount++;

		if (getConsideredStateStore().contains(newRec)) {
			// The state at index was visited before, hence the estimate is
			// irrelevant and we skip
			newRec.setEstimatedRemainingCost(ESTIMATEIRRELEVANT, true);
			traverseEdge(rec, newRec);
			return;
		}

		// We need to compute the data cost
		newRec = augmentRecord(newRec);
		// Skip all hopeless states
		if (newRec == null) {
			return;
		}

		final int c;
		if (isRandomizedQueue) {
			// we want to add records to the queue even if they are already there 
			c = -1;
		} else {
			Record r = queue.contains(newRec);
			// if r!=null then there is a record in the queue pointing to this
			// state.			
			if (r != null) {
				// We reached the state at index before, check the costs and
				// get the estimate from the previous time. Note that we CANNOT
				// get a new estimate from the tail since this would invalidate
				// the
				// ordering relation in the priority queue
				c = r.getCostSoFar();
				newRec.setEstimatedRemainingCost(r.getEstimatedRemainingCost(), true);
			} else {
				// state is not already in the queue				
				c = -1;
			}
		}

		if (c >= 0 && c <= newRec.getCostSoFar()) {
			// A record with the same state and lower cost exists.
			if (stateIndex >= 0) {
				traverseEdge(rec, newRec);
			}
			return;
		}

		final T newTail;
		if (stateIndex >= 0) {
			// so far, we did the cheap stuff, now get the new tail from
			// storage, knowing that
			// a state exists at index
			int h;
			try {
				h = storageHandler.getEstimate(newHead, stateIndex);
			} catch (Exception e) {
				throw new AStarException(e);
			}

			newRec.setState(stateIndex);
			newRec.setEstimatedRemainingCost(h, true);

			if (queue.add(newRec)) {
				queuedStateCount++;
			}

			traverseEdge(rec, newRec);
			return;
		}

		// the state space doesn't contain a corresponding state, hence we need
		// to compute the tail.
		newTail = computeNewTail(rec, tail, newHead, modelMove, movedEvent, activity);

		if (!newTail.canComplete()) {
			return;
		}

		int h = newTail.getEstimatedCosts(delegate, newHead);

		// Check if the head is in the store and add if it isn't.
		final State<H, T> newState = new State<>(newHead, newTail);

		try {
			storageHandler.storeStateForRecord(newState, newRec);
		} catch (Exception e) {
			throw new AStarException(e);
		}

		newRec.setEstimatedRemainingCost(h, true);
		if (queue.add(newRec)) {
			queuedStateCount++;
		}

		traverseEdge(rec, newRec);

	}

	private void traverseEdge(Record rec, DRecord newRec) {
		// edge from oldRec to newRec traversed.
		for (AStarObserver observer : observers) {
			observer.edgeTraversed(rec, newRec);
		}
	}

}