package org.processmining.plugins.balancedconformance.controlflow.override.storage;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.LockSupport;

import nl.tue.astar.AStarException;
import nl.tue.astar.Delegate;
import nl.tue.astar.FastLowerBoundTail;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.impl.AbstractAStarThread;
import nl.tue.astar.impl.State;

public final class JavaCollectionStorageHandlerConcurrent<H extends Head, T extends Tail> implements
		AbstractAStarThread.StorageHandler<H, T> {

	private static final Integer INSERTING = new Integer(Integer.MIN_VALUE);

	private final AtomicInteger keyCounter;
	
	protected final ConcurrentMap<H, Integer> keyMap;
	protected final ConcurrentMap<Integer, State<H, T>> reverseMap;
	private final Delegate<H, T> delegate;

	public JavaCollectionStorageHandlerConcurrent(Delegate<H, T> delegate) {
		this.delegate = delegate;
		this.keyMap = new ConcurrentHashMap<>();
		this.reverseMap = new ConcurrentHashMap<>();
		this.keyCounter = new AtomicInteger(-1);
	}

	public T getStoredTail(T tail, long index, int modelMove, int movedEvent, int logMove) {
		return internalGet((int) index).getTail();
	}

	private State<H, T> internalGet(int index) {
		return reverseMap.get(index);
	}

	public void storeStateForRecord(State<H, T> val, Record newRec) {
		// First check for existence
		Integer index = keyMap.get(val.getHead());

		if (index != null) {
			while (index == INSERTING) {
				LockSupport.parkNanos(1L);
				index = keyMap.get(val.getHead());
			}
			// Already inserted
			newRec.setState(index); 
			return; 
		} else {
			// No yet present, insert a stub value as we don't want to waste indices
			index = keyMap.putIfAbsent(val.getHead(), INSERTING);
			if (index == null) {
				// We reserved the spot and, therefore, the following operations are executed only once for each key
				// Take care that this block never fails, otherwise other threads trying the same put starve
				index = keyCounter.incrementAndGet();
				reverseMap.put(index, val);
				keyMap.put(val.getHead(), index);
				newRec.setState(index);
				return;
			} else {
				// Another thread reserved the sport, wait until insert is visible to us
				while (index == INSERTING) {
					LockSupport.parkNanos(1L);
					index = keyMap.get(val);
				}
				newRec.setState(index);
				return;
			}
		}
	}

	public long getIndexOf(H head) {
		Integer index = keyMap.get(head);
		while (index != null && index == INSERTING) {
			LockSupport.parkNanos(1L);
			index = keyMap.get(head);
		}
		return index != null ? index.longValue() : -1;
	}

	public State<H, T> getStoredState(Record rec) {
		return internalGet((int) rec.getState());
	}

	@Override
	public int getEstimate(H head, long index) {
		T tail = internalGet((int) index).getTail();
		if (tail instanceof FastLowerBoundTail) {
			if (!((FastLowerBoundTail) tail).isExactEstimateKnown()) {
				return -tail.getEstimatedCosts(delegate, head) - 1;
			}
		}
		return tail.getEstimatedCosts(delegate, head);
	}

	@Override
	public <TT extends FastLowerBoundTail> void reComputeFastLowerboundTail(Record rec, TT tail, H head,
			int storedEstimate) throws AStarException {
		tail.computeEstimate(delegate, head, storedEstimate);
	}
}
