/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow;

import java.util.ArrayList;
import java.util.List;

import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

/**
 * {@link DataAlignmentAdapter} that calls multiple other adapters in order and
 * returns the first valid result.
 * 
 * @author F. Mannhardt
 * 
 */
public class DataAlignmentAdapterChainingImpl implements DataAlignmentAdapter {

	private static final boolean DEBUG = false;

	private final DataStateFactory dataStateFactory;
	private final List<DataAlignmentAdapter> adapterList;
	private final List<PartialDataAlignmentObserver> observerList;

	private int partialDataAlignments = 0;

	public DataAlignmentAdapterChainingImpl(DataStateFactory dataStateFactory) {
		super();
		this.dataStateFactory = dataStateFactory;
		this.adapterList = new ArrayList<>();
		this.observerList = new ArrayList<>();
	}

	@Override
	public DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowResult) throws DataAlignmentException {
		throw new UnsupportedOperationException(
				"This DataAlignment adapter does not support a complete data alignment!");
	}

	@Override
	public PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
			PartialResultFactory partialResultBuilder, DRecord currentRecord) throws DataAlignmentException {

		partialResultBuilder.newRecord();

		for (int i = 0; i < adapterList.size(); i++) {
			DataAlignmentAdapter adapter = adapterList.get(i);
			if (adapter instanceof ExpensiveAdapter && !(adapter instanceof DataAlignmentAdapterWithoutPartialImpl)) {
				partialDataAlignments++;
			}
			PartialDataAlignedTrace result = adapter.retrievePartialDataAlignment(traceMapping, partialResultBuilder,
					currentRecord);

			assert testWithLastInChain(traceMapping, currentRecord, partialResultBuilder, adapter,
					result) : "Invalid result from cache or optimization. Look at the console for more information.";

			if (result != PartialDataAlignedTrace.NO_RESULT) {
				for (int j = 0; j < observerList.size(); j++) {
					observerList.get(j).newPartialDataAlignment(currentRecord, result);
				}
				if (DEBUG) {
					if (result == PartialDataAlignedTrace.IMPOSSIBLE) {
						System.out.println(
								"IMPOSSIBLE: " + partialResultBuilder.buildPartialFromCurrentRecord(currentRecord));
					} else {
						System.out.println(result + " (" + currentRecord.getState() + "): "
								+ partialResultBuilder.buildPartialFromCurrentRecord(currentRecord));
					}
				}
				return result;
			}
		}
		if (DEBUG) {
			System.out.println("NO RESULT: " + partialResultBuilder.buildPartialFromCurrentRecord(currentRecord));
		}

		return PartialDataAlignedTrace.NO_RESULT;
	}

	private boolean testWithLastInChain(TraceMapping traceMapping, DRecord currentRecord,
			PartialResultFactory partialResultBuilder, DataAlignmentAdapter adapter, PartialDataAlignedTrace result)
			throws DataAlignmentException {

		if (result != PartialDataAlignedTrace.NO_RESULT && result != PartialDataAlignedTrace.IMPOSSIBLE
				&& (!(adapter instanceof ExpensiveAdapter))) {
			DataAlignmentAdapter ilpAdapter = adapterList.get(adapterList.size() - 1);
			PartialDataAlignedTrace ilpResult = ilpAdapter.retrievePartialDataAlignment(traceMapping,
					partialResultBuilder, currentRecord);

			if (ilpResult.getDataCost() != result.getDataCost()
					&& !ilpResult.getVariableAssignment().equals(result.getVariableAssignment())) {
				System.out.println("ILP returned different result " + ilpResult.getDataCost() + " vs. "
						+ result.getDataCost() + " by " + adapter.getClass().getSimpleName() + " for "
						+ partialResultBuilder.buildPartialFromCurrentRecord(currentRecord));
				System.out.println("ILP returned different result " + ilpResult.getVariableAssignment() + " vs. "
						+ result.getVariableAssignment() + " by " + adapter.getClass().getSimpleName() + " for "
						+ partialResultBuilder.buildPartialFromCurrentRecord(currentRecord));
				return false;
			} else {
				return true;
			}

		} else {
			return true;
		}
	}

	public DataAlignmentAdapterChainingImpl addAdapter(DataAlignmentAdapter adapter) {
		adapterList.add(adapter);
		return this;
	}

	public DataAlignmentAdapterChainingImpl addObserver(PartialDataAlignmentObserver observer) {
		observerList.add(observer);
		return this;
	}

	public int getPartialDataAlignments() {
		return partialDataAlignments;
	}

	public DataStateStore newDataStateStore(int initialCapacity) {
		throw new UnsupportedOperationException("Data Chaining Adapter does not support newDataStateStore");
	}

	public DataStateFactory getDataStateFactory() {
		return dataStateFactory;
	}

	public void close() {
	}

}