package org.processmining.plugins.balancedconformance.dataflow.lp.lpsolve;

import java.util.Map;

import org.processmining.plugins.balancedconformance.dataflow.lp.lpsolve.DataAlignmentSolverFactoryLpSolve.LpSolveInstance;

import com.google.common.base.Throwables;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;

import lpsolve.LpSolve;
import lpsolve.LpSolveException;
import net.sf.javailp.Constraint;
import net.sf.javailp.Linear;
import net.sf.javailp.OptType;
import net.sf.javailp.Problem;
import net.sf.javailp.Result;
import net.sf.javailp.ResultImpl;
import net.sf.javailp.SolverLpSolve;
import net.sf.javailp.Term;
import net.sf.javailp.VarType;

public final class DataAlignmentLpSolveSolver extends SolverLpSolve {

	private final DataAlignmentSolverFactoryLpSolve solverFactory;

	public DataAlignmentLpSolveSolver(DataAlignmentSolverFactoryLpSolve solverFactory) {
		this.solverFactory = solverFactory;
	}

	@Override
	public Result solve(Problem problem) {
		try {
			BiMap<Object, Integer> varToIndex = getVariableMapping(problem);

			LpSolveInstance lpInstance = solverFactory.getLp();
			try {
				LpSolve lp = lpInstance.lp;

				lp.resizeLp(problem.getConstraintsCount(), problem.getVariablesCount());
				for (int j = 0; j < problem.getVariablesCount(); j++) {
					lp.addColumn(new double[problem.getConstraintsCount()]);

				}

				lp.setAddRowmode(true);

				if (problem.getObjective() != null) {
					addObjective(problem, varToIndex, lp);	
				}				
				addConstraints(problem, varToIndex, lp);

				lp.setAddRowmode(false);

				setVariableBounds(problem, varToIndex, lp);
						
				int ret = lp.solve();

				// 0 means optimal
				// 1 means suboptimal
				// 12 means feasible
				if (ret != 0 && ret != 1 && ret != 12) {
					return null;
				}
				
				return retrieveResult(problem, varToIndex, lp);

			} finally {				
				solverFactory.returnLp(lpInstance);
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return null;
		} catch (LpSolveException e) {
			System.err.println("LpSolve Excetpion, Cause: " + Throwables.getStackTraceAsString(e));
			return null;
		}
	}

	private Result retrieveResult(Problem problem, BiMap<Object, Integer> varToIndex, LpSolve lp) throws LpSolveException {
		final Result result;
		if (problem.getObjective() != null) {
			result = new ResultImpl(problem.getObjective());
		} else {
			result = new ResultImpl();
		}
		
		double[] values = new double[problem.getVariablesCount()];
		lp.getVariables(values); 

		for (Object variable : problem.getVariables()) {

			int index = varToIndex.get(variable);
			VarType varType = problem.getVarType(variable);

			double value = values[index-1];
			// does not work since lpsolve messes up with presolve 
			//double value = lp.getVarPrimalresult(lp.getNorigRows()+index);			
			
			if (varType == VarType.INT || varType == VarType.BOOL) {
				int v = (int) Math.round(value);
				result.putPrimalValue(variable, v);
			} else {
				result.putPrimalValue(variable, value);
			}
		}
		return result;
	}

	private void setVariableBounds(Problem problem, BiMap<Object, Integer> varToIndex, LpSolve lp) throws LpSolveException {
		for (Object variable : problem.getVariables()) {
			int index = varToIndex.get(variable);

			VarType varType = problem.getVarType(variable);
			Number lowerBound = problem.getVarLowerBound(variable);
			Number upperBound = problem.getVarUpperBound(variable);

			if (varType == VarType.INT) {
				lp.setInt(index, true);
			} else if (varType == VarType.BOOL) {
				lp.setBinary(index, true);
			}

			if (varType != VarType.BOOL) {
				if (lowerBound != null) {
					lp.setLowbo(index, lowerBound.doubleValue());
				}
				if (upperBound != null) {
					lp.setUpbo(index, upperBound.doubleValue());
				}
			}

		}
	}

	private void addConstraints(Problem problem, BiMap<Object, Integer> varToIndex, LpSolve lp) throws LpSolveException {
		for (Constraint constraint : problem.getConstraints()) {
			int size = constraint.size();

			int[] var = new int[size];
			double[] coeffs = new double[size];
			Linear linear = constraint.getLhs();

			convert(linear, var, coeffs, varToIndex);

			int operator;
			switch (constraint.getOperator()) {
				case LE :
					operator = LpSolve.LE;
					break;
				case GE :
					operator = LpSolve.GE;
					break;
				default : // EQ
					operator = LpSolve.EQ;
			}

			double rhs = constraint.getRhs().doubleValue();
			lp.addConstraintex(size, coeffs, var, operator, rhs);
		}
	}

	private void addObjective(Problem problem, BiMap<Object, Integer> varToIndex, LpSolve lp) throws LpSolveException {
		Linear objective = problem.getObjective();
		int size = objective.size();
		int[] var = new int[size];
		double[] coeffs = new double[size];

		convert(objective, var, coeffs, varToIndex);

		lp.setObjFnex(size, coeffs, var);

		if (problem.getOptType() != OptType.MIN) {
			lp.setMaxim();
		}
	}

	private BiMap<Object, Integer> getVariableMapping(Problem problem) {
		BiMap<Object, Integer> varToIndex = HashBiMap.create(problem.getVariablesCount());
		int i = 1;
		for (Object variable : problem.getVariables()) {
			varToIndex.put(variable, i++);
		}
		return varToIndex;
	}

	@Override
	protected void convert(Linear linear, int[] var, double[] coeffs, Map<Object, Integer> varToIndex) {
		for (int i = 0; i < linear.size(); i++) {
			Term term = linear.get(i);
			var[i] = varToIndex.get(term.getVariable());
			coeffs[i] = term.getCoefficient().doubleValue();
		}
	}

}
