/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.mapping;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeTimestampImpl;
import org.processmining.datapetrinets.exception.VariableNotFoundException;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.DataConformance.Utility;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult.Step;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable.Prefix;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable.PrefixStep;

import com.google.common.base.Function;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.Iterables;

/**
 * Contains information about the trace
 * 
 * @author F. Mannhardt
 * 
 */
public final class TraceMapping {

	private static final class PrefixStepImpl implements PrefixStep {

		private final Transition transition;
		private final XEvent event;

		public PrefixStepImpl(Transition transition, XEvent event) {
			this.transition = transition;
			this.event = event;
		}

		public Transition getProcessObject() {
			return transition;
		}

		public XEvent getLogObject() {
			return event;
		}

		public String toString() {
			return MessageFormat.format("({0},{1})", transition != null ? transition.getLabel() : "N/A",
					event != null ? XUtils.getConceptName(event) : "N/A");
		}

	}

	private static final Function<Step, PrefixStep> STEP_TO_PREFIX = new Function<Step, PrefixStep>() {

		public PrefixStep apply(final Step step) {
			return new PrefixStepImpl(step.getModelNode(), step.getLogNode());
		}
	};

	private static final class AlignmentPrefixLazyImpl implements Prefix {

		private final TraceMapping traceMapping;
		private final PartialResultFactory partialResultFactory;
		private final DRecord record;
		private final DataStateStore store;
		private DataState previousDataState;

		public AlignmentPrefixLazyImpl(TraceMapping traceMapping, PartialResultFactory partialResultFactory,
				DRecord record, DataStateStore store) {
			this.traceMapping = traceMapping;
			this.record = record;
			this.partialResultFactory = partialResultFactory;
			this.store = store;
		}

		public Iterable<PrefixStep> getSteps() {
			return Iterables.transform(partialResultFactory.buildPartialFromCurrentRecord(record).getSteps(),
					STEP_TO_PREFIX);
		}

		public Object variableValue(String variableName) {
			if (record.getPredecessor() != null) {
				if (previousDataState == null) {
					previousDataState = record.getPredecessor().getDataState(store);
				}
				if (previousDataState != null) {
					Variable variable = traceMapping.getLogMapping().getVariables().get(variableName);
					if (variable != null) {
						return DataStateHelper.tryGetPrimeOrNormalObjectFromDataState(previousDataState, variable
								.getName(), variable.getType(), traceMapping.getLogMapping().getStringDiscretizer());
					} else {
						throw new VariableNotFoundException("Variable " + variableName + " is unkown!");
					}
				}
			}
			return null;
		}

		public boolean isAligned() {
			return true;
		}

	}

	private static final Function<XEvent, PrefixStep> EVENT_TO_PREFIX = new Function<XEvent, PrefixStep>() {

		public PrefixStep apply(XEvent event) {
			return new PrefixStepImpl(null, event);
		}
	};

	private static final class MappedTracePrefixImpl implements Prefix {

		private final List<XEvent> mappedPrefix;

		public MappedTracePrefixImpl(List<XEvent> mappedPrefix) {
			this.mappedPrefix = mappedPrefix;
		}

		public Iterable<PrefixStep> getSteps() {
			return Iterables.transform(mappedPrefix, EVENT_TO_PREFIX);
		}

		public Object variableValue(String variableName) {
			return null;
		}

		public boolean isAligned() {
			return true;
		}

	}

	private static final class MappedEvent {

		private XEvent event;
		private Transition transition;

		public MappedEvent() {
			super();
		}

		public MappedEvent(XEvent event, Transition transition) {
			super();
			this.event = event;
			this.transition = transition;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((event == null) ? 0 : event.hashCode());
			result = prime * result + ((transition == null) ? 0 : transition.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof MappedEvent))
				return false;
			MappedEvent other = (MappedEvent) obj;
			if (event == null) {
				if (other.event != null)
					return false;
			} else if (!event.equals(other.event))
				return false;
			if (transition == null) {
				if (other.transition != null)
					return false;
			} else if (!transition.equals(other.transition))
				return false;
			return true;
		}

		public void setEvent(XEvent event) {
			this.event = event;
		}

		public void setTransition(Transition transition) {
			this.transition = transition;
		}

	}

	private final LogMapping logMapping;

	private final String traceName;

	/**
	 * The original trace
	 */
	private final XTrace trace;

	/**
	 * A view on the trace containing only the mapped events
	 */
	private final List<XEvent> mappedTrace;

	/**
	 * Assignments by the events & virtual variables that can be computed
	 * without the alignment prefix
	 */
	private final ImmutableMap<MappedEvent, Map<String, Object>> event2VariablesAssignments;

	/**
	 * Creates {@link ControlFlowAlignmentResult} from a {@link DRecord}
	 */
	private PartialResultFactory partialResultFactory;

	private DataStateStore dataStateStore;

	public TraceMapping(XTrace trace, LogMapping logMapping) throws DataAlignmentException {
		super();
		this.logMapping = logMapping;
		String name = XUtils.getConceptName(trace);
		this.traceName = name != null ? name : UUID.randomUUID().toString();
		this.trace = trace;
		Date startTime = findStartTimeOfMappedTrace(logMapping, trace);
		this.mappedTrace = new ArrayList<>();		
		for (XEvent event : trace) {
			if (!logMapping.getMappedTransitions(event).isEmpty()) {
				if (logMapping.hasVirtualVariables() && !(event.getAttributes().containsKey(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME))) {
					Date currentTime = XUtils.getTimestamp(event);
					if (currentTime != null && startTime != null) {
						XUtils.putAttribute(
								event,
								new XAttributeTimestampImpl(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME, currentTime
										.getTime() - startTime.getTime()));
					}
				}
				mappedTrace.add(event);
			}
		}
		this.event2VariablesAssignments = buildVariableAssignmentMap(mappedTrace, logMapping);
	}

	private static Date findStartTimeOfMappedTrace(LogMapping logMapping, XTrace trace) {
		for (XEvent event : trace) {
			if (!logMapping.getMappedTransitions(event).isEmpty()) {
				Date startTime = XUtils.getTimestamp(event);
				if (startTime != null) {
					return startTime;
				}
			}
		}
		// trace without time information
		return null;
	}

	private static ImmutableMap<MappedEvent, Map<String, Object>> buildVariableAssignmentMap(List<XEvent> mappedTrace,
			LogMapping mapping) throws DataAlignmentException {
		Builder<MappedEvent, Map<String, Object>> builder = ImmutableMap.builder();
		for (ListIterator<XEvent> iterator = mappedTrace.listIterator(); iterator.hasNext();) {
			XEvent e = iterator.next();
			// An event class can be mapped to multiple transitions
			final Set<Transition> mappedTransitions = mapping.getMappedTransitions(e);
			for (Transition transition : mappedTransitions) {
				// Each XEvent is unique
				Map<String, Object> variableAssignment = getVariableAssignmentFromEventLog(
						mappedTrace.subList(0, iterator.previousIndex()), e, transition, mapping);
				if (variableAssignment != null) {
					builder.put(new MappedEvent(e, transition), variableAssignment);
				}
			}
		}
		return builder.build();
	}

	private static ImmutableMap<String, Object> getVariableAssignmentFromEventLog(List<XEvent> prefix, XEvent event,
			Transition transition, LogMapping logMapping) throws DataAlignmentException {
		Builder<String, Object> builder = null;
		if (event != null) {
			Set<String> variablesToWrite = logMapping.getVariablesToWrite().get(transition);
			assert variablesToWrite != null : "Variables to write is null for " + transition;

			for (String variableName : variablesToWrite) {
				Variable variable = logMapping.getVariables().get(variableName);

				// Virtual variables that don't need a prefix can be pre-processed!
				if (variable.isVirtual() && !variable.getVirtualVariable().needsAlignedPrefix()) {

					Object computedValue = variable.getVirtualVariable().compute(new MappedTracePrefixImpl(prefix),
							transition, event, logMapping);
					if (computedValue != null) {
						builder = lazyInitBuilder(builder);
						builder.put(variableName, computedValue);
					}

				} else if (!variable.isVirtual()) {

					String attributeName = variable.getAttributeName();
					XAttribute attribute = event.getAttributes().get(variable.getAttributeName());
					if (attribute != null) {
						Object value = org.processmining.plugins.DataConformance.Utility.getValue(attribute);
						Class<?> expectedType = variable.getType();
						if (Utility.isAssignableToExpectedType(value, expectedType)) {
							builder = lazyInitBuilder(builder);
							builder.put(variableName, Utility.convertToExpectedType(value, expectedType));
						} else {
							throw new DataAlignmentException(
									"Reading event "
											+ logMapping.getEventClassifier().getClassIdentity(event)
											+ " failed. Attribute "
											+ attributeName
											+ " with value "
											+ value
											+ " is not of expected type "
											+ expectedType.getSimpleName()
											+ " for variable: "
											+ variableName
											+ ". Check if the type of your variables in the DPN-net matches the type of the attributes in the XES event. "
											+ " Please notice that for XES attributes <int> is converted to java.lang.Long and <float> is converted to java.lang.Double. "
											+ "This plugin will not silently convert a <float> attribute to java.lang.Double, as this may result in a loss of precision! "
											+ "You may have to change the variables of your DPN-net to java.lang.Long and java.lang.Double instead of java.lang.Integer and java.lang.Float!");
						}
					}
				}
			}
		}

		if (builder == null) {
			return ImmutableMap.of();
		} else {
			return builder.build();
		}
	}

	private static Builder<String, Object> lazyInitBuilder(Builder<String, Object> builder) {
		if (builder == null) {
			builder = ImmutableMap.builder();
		}
		return builder;
	}

	private final MappedEvent mappedEvent = new MappedEvent();

	private void prepareMappedEvent(final XEvent currentEvent, final Transition currentTransition) {
		mappedEvent.setEvent(currentEvent);
		mappedEvent.setTransition(currentTransition);
	}

	public Object getVariableAssignmentFor(String variableName, final DRecord currentRecord, final XEvent currentEvent,
			final Transition currentTransition) {

		final Object value = getValueFromEvent(variableName, currentEvent, currentTransition);
		if (value != null) {
			return value;
		} else {
			return getVirtualVariableAssignmentFor(variableName, currentRecord, currentEvent, currentTransition);
		}

	}

	private Object getValueFromEvent(String variableName, XEvent currentEvent, Transition currentTransition) {
		prepareMappedEvent(currentEvent, currentTransition);
		Map<String, Object> assignments = event2VariablesAssignments.get(mappedEvent);
		Object value = null;
		if (assignments != null) {
			value = assignments.get(variableName);
		}
		return value;
	}

	private Object getVirtualVariableAssignmentFor(String variableName, final DRecord currentRecord,
			final XEvent currentEvent, final Transition currentTransition) {
		if (logMapping.hasVirtualVariables()) {
			Variable variable = logMapping.getVariables().get(variableName);
			if (variable.isVirtual()) {
				Prefix alignmentPrefix = new AlignmentPrefixLazyImpl(this, getPartialResultFactory(), currentRecord,
						dataStateStore);
				return variable.getVirtualVariable().compute(alignmentPrefix, currentTransition, currentEvent,
						logMapping);
			} else {
				// Missing assignment
				return null;
			}
		} else {
			// Missing assignment
			return null;
		}
	}

	public XTrace getTrace() {
		return trace;
	}

	public List<XEvent> getMappedTrace() {
		return mappedTrace;
	}

	public String getTraceName() {
		return traceName;
	}

	public LogMapping getLogMapping() {
		return logMapping;
	}

	public PartialResultFactory getPartialResultFactory() {
		return partialResultFactory;
	}

	public void setPartialResultFactory(PartialResultFactory partialResultFactory) {
		this.partialResultFactory = partialResultFactory;
	}

	public DataStateStore getDataStateStore() {
		return dataStateStore;
	}

	public void setDataStateStore(DataStateStore dataStateStore) {
		this.dataStateStore = dataStateStore;
	}

}