/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.ui;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.util.ui.widgets.ProMHeaderPanel;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;


/**
 * @author F. Mannhardt
 *
 */
public class CheckVariableBoundsPanel extends ProMHeaderPanel
{
	
	private static final int ROUND_TO = 1000;

	private class CellValueVerification extends DefaultCellEditor
	{
		private static final long serialVersionUID = 6367444400714673476L;
		
		private JTable table;
		private JTextField field=new JTextField();

		public CellValueVerification(JTable table)
		{
			this(table,new JTextField());
		}
		
		private CellValueVerification(JTable table,JTextField aField) {
			super(aField);
			field=aField;
			this.table=table;
		}
		
		@Override
		public Object getCellEditorValue() {
			String value=field.getText();
			Class<?> varType= (Class<?>)table.getModel().getValueAt(table.getEditingRow(), 1);
			
			try {
				if (varType == String.class) {
					return value;
				} else if (varType == Long.class) {
					Long.parseLong(value);
					return value;
				} else if (varType == Integer.class) {
					Integer.parseInt(value);
					return value;
				} else if (varType == Double.class) {
					Double.parseDouble(value);
					return value;				
				} else if (varType == Float.class) {
					Float.parseFloat(value);
					return value;
				} else if (varType == Date.class) {
					new Date(Long.parseLong(value));
					return value;				
				} else {
					return value;
				}
				
			} catch (NumberFormatException e) {
				return "";
			}

		}
		
	}
	
	private static final long serialVersionUID = 2L;
	
	private Object columnNames[] = { "Variable Name", "Type", "Minimum Value", "Maximum Value"};
	private DefaultTableModel dtm=new DefaultTableModel(columnNames, 0);	
	private JTable table=new JTable(dtm);
	
	public CheckVariableBoundsPanel(DataPetriNet inputNet, XLog log, Map<String, String> variableMapping) {
		super("Verify/Set Process Variable Bounds");
		init();
		DataElement[] variables=inputNet.getVariables().toArray(new DataElement[0]);
		Arrays.sort(variables, new Comparator<DataElement>() {

			public int compare(DataElement o1, DataElement o2) {
				return o1.getVarName().compareTo(o2.getVarName());
			}
		});
		
		
		Map<DataElement, Number> minValueMap = new HashMap<>();
		Map<DataElement, Number> maxValueMap = new HashMap<>();

		for (XTrace trace: log) {
			for (XEvent event: trace) {
				
				XAttributeMap attributes = event.getAttributes();

				for(DataElement elem : variables) {
					
					XAttribute attribute = attributes.get(variableMapping.get(elem.getVarName()));
					Number minValue = minValueMap.get(elem);
					Number maxValue = maxValueMap.get(elem);					
				
					if (attribute != null) {
						if (attribute instanceof XAttributeDiscrete) {
							XAttributeDiscrete dAttr = (XAttributeDiscrete) attribute;
							minValueMap.put(elem, minValue == null ? dAttr.getValue() : Math.min(minValue.longValue(), dAttr.getValue()));
							maxValueMap.put(elem, maxValue == null ? dAttr.getValue() : Math.max(maxValue.longValue(), dAttr.getValue()));														
						} else if (attribute instanceof XAttributeContinuous) {
							XAttributeContinuous cAttr = (XAttributeContinuous) attribute;
							minValueMap.put(elem, minValue == null ? cAttr.getValue() : Math.min(minValue.doubleValue(), cAttr.getValue()));
							maxValueMap.put(elem, maxValue == null ? cAttr.getValue() : Math.max(maxValue.doubleValue(), cAttr.getValue()));
						} else if (attribute instanceof XAttributeTimestamp) {
							XAttributeTimestamp tAttr = (XAttributeTimestamp) attribute;
							minValueMap.put(elem, minValue == null ? tAttr.getValueMillis() : Math.min(minValue.doubleValue(), tAttr.getValueMillis()));
							maxValueMap.put(elem, maxValue == null ? tAttr.getValueMillis() : Math.max(maxValue.doubleValue(), tAttr.getValueMillis()));							
						}
					}
					
				}
			}
		}
		
		
		for(DataElement elem : variables) {
			
			Object row[]=new Object[4];
			row[0]=elem.getVarName();
			row[1]=elem.getType();

			Number minValue= determineMinValue(elem, minValueMap);
			Number maxValue= determineMaxValue(elem, maxValueMap);
			
			if (minValue != null && minValue.equals(maxValue)) {
				maxValue = minValue.longValue() + ROUND_TO;
			}
			
			if (minValue!=null) {
				if (elem.getType() == Date.class) {
					row[2] = minValue.toString();
				} else {
					row[2] = minValue.toString();	
				}				
			} else {
				row[2]="";
			}
			
			if (maxValue!=null) {
				row[3]=maxValue.toString();
			} else {
				row[3]="";
			}
			dtm.addRow(row);
			
		}
		
		init();
	}

	private Number determineMaxValue(DataElement elem, Map<DataElement, Number> maxValueMap) {
		if (elem.getMaxValue() != null && elem.getMaxValue() instanceof Number) {
			return (Number) elem.getMaxValue();
		} else {
			Number maxValue = maxValueMap.get(elem);
			if (maxValue != null) {
				double roundedValue = Math.ceil((maxValue.doubleValue()/ROUND_TO))*ROUND_TO;
				if (elem.getType() == Long.class) {
					return new Long((long)roundedValue);
				} else if (elem.getType() == Integer.class) {
					return new Integer((int)roundedValue);
				} else if (elem.getType() == Double.class) {
					return new Double(roundedValue);
				} else if (elem.getType() == Float.class) {
					return new Float(roundedValue);
				} else if (elem.getType() == Date.class) {
					return new Long((long)roundedValue);
				} else {
					return null;
				}
			} else {
				return null;
			}
		}
	}

	private Number determineMinValue(DataElement elem, Map<DataElement, Number> minValueMap) {
		if (elem.getMinValue() != null && elem.getMinValue() instanceof Number) {			
			return (Number) elem.getMinValue();
		} else {
			Number minValue = minValueMap.get(elem);
			if (minValue != null) {
				double roundedValue = Math.floor((minValue.doubleValue()/ROUND_TO))*ROUND_TO;
				if (elem.getType() == Long.class) {
					return new Long((long)roundedValue);
				} else if (elem.getType() == Integer.class) {
					return new Integer((int)roundedValue);
				} else if (elem.getType() == Double.class) {
					return new Double(roundedValue);
				} else if (elem.getType() == Float.class) {
					return new Float(roundedValue);
				} else if (elem.getType() == Date.class) {
					return new Long((long)roundedValue);
				} else {
					return null;
				}
			} else {
				return null;
			}
		}
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	private final void init()
	{				
		JComboBox combo = new JComboBox();
		combo.addItem(String.class);
		combo.addItem(Integer.class);
		combo.addItem(Float.class);
		combo.addItem(Date.class);
		combo.addItem(Boolean.class);
		table.getColumnModel().getColumn(3).setCellEditor(new CellValueVerification(table));
		table.getColumnModel().getColumn(2).setCellEditor(new CellValueVerification(table));
		table.getColumnModel().getColumn(1).setCellEditor(new DefaultCellEditor(combo));
		
		this.setLayout(new BorderLayout());
		this.add(new JScrollPane(table), BorderLayout.CENTER);
		JPanel south=new JPanel(new FlowLayout(FlowLayout.CENTER));
		this.add(south,BorderLayout.SOUTH);		
	}

	public Set<DataElement> getDataElements()
	{
		int index=dtm.getRowCount()-1;
		Set<DataElement> retValue = new HashSet<>();
		for(;index>=0;index--)
		{
			DataElement element;
			String varName=(String) dtm.getValueAt(index, 0);
			Class<?> varType=(Class<?>) dtm.getValueAt(index, 1);
			String minValueString=(String)dtm.getValueAt(index, 2);
			if (minValueString!=null && minValueString.equals(""))
				minValueString=null;
			String maxValueString=(String)dtm.getValueAt(index, 3);
			if (maxValueString!=null && maxValueString.equals(""))
				maxValueString=null;
			
			
			Comparable<?> minValue=null;
			Comparable<?> maxValue=null;
			
			try {
				if (varType == String.class) {
					minValue = minValueString;
					maxValue = maxValueString;
				} else if (varType == Long.class) {
					minValue = minValueString != null ? Long.parseLong(minValueString) : null;
					maxValue = maxValueString != null ? Long.parseLong(maxValueString) : null;
				} else if (varType == Integer.class) {
					minValue = minValueString != null ? Integer.parseInt(minValueString) : null;
					maxValue = maxValueString != null ? Integer.parseInt(maxValueString) : null;				
				} else if (varType == Double.class) {
					minValue = minValueString != null ? Double.parseDouble(minValueString) : null;
					maxValue = maxValueString != null ? Double.parseDouble(maxValueString) : null;				
				} else if (varType == Float.class) {
					minValue = minValueString != null ? Float.parseFloat(minValueString) : null;
					maxValue = maxValueString != null ? Float.parseFloat(maxValueString) : null;
				} else if (varType == Date.class) {
					minValue = minValueString != null ? new Date(Long.parseLong(minValueString)) : null;
					maxValue = maxValueString != null ? new Date(Long.parseLong(maxValueString)) : null;				
				}
			} catch (NumberFormatException e) {
			} 
			
			element = new DataElement(varName, varType, minValue, maxValue, null);
			retValue.add(element);
		}
		return retValue;
		
	}
	
}