package org.processmining.plugins.balancedconformance.visualisation;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Iterator;
import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult.Step;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResultImpl;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

import com.google.common.collect.Iterables;

import it.unimi.dsi.fastutil.Hash.Strategy;
import it.unimi.dsi.fastutil.objects.Object2LongOpenCustomHashMap;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Head;
import nl.tue.astar.ObservableAStarThread;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.DotGraphAStarObserver;

/**
 * Adds abbreviated event/transition names to the search graph that is saved to
 * a DOT file. Derived from original {@link DotGraphAStarObserver}.
 * 
 * @author fmannhardt
 * 
 */
public class DotDataAStarObserver implements AStarObserver {

	protected Writer writer;
	protected int visitCounter;
	protected int finalNodeCounter;
	protected final File outputFile;

	private final AbstractPDelegate<? extends Tail> delegate;
	private final Map<XEvent, XEventClass> mapEv2EventClass;
	private final Map<XEventClass, Integer> mapEvClass2Cost;
	private final TraceMapping traceMapping;
	private final Trace trace;
	
	private final Object2LongOpenCustomHashMap<DRecord> stateMap;

	public DotDataAStarObserver(File outputFile, AbstractPDelegate<? extends Tail> delegate,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost,
			TraceMapping traceMapping, Trace trace, Strategy<DRecord> strategy) {
		super();
		this.outputFile = outputFile;
		this.delegate = delegate;
		this.mapEv2EventClass = mapEv2EventClass;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.traceMapping = traceMapping;
		this.trace = trace;
		this.stateMap = new Object2LongOpenCustomHashMap<>(strategy);
	}

	private ControlFlowAlignmentResult calcCFAlignmentResult(Record record) {
		return new ControlFlowAlignmentResultImpl(((DRecord) record), delegate, mapEv2EventClass, mapEvClass2Cost,
				traceMapping.getTrace(), traceMapping.getMappedTrace(), trace, true);
	}

	private String buildLastStepString(ControlFlowAlignmentResult result) {
		StringBuilder sb = new StringBuilder();
		Step step = Iterables.getLast(result.getSteps());
		switch (step.getType()) {
			case L :
				sb.append("(");
				sb.append(getStepTypeStart(step.getType()));
				sb.append(getNodeAbbrv(step.getLogNode()));
				sb.append(getStepTypeEnd(step.getType()));
				sb.append(",");
				sb.append("&gt;&gt;");
				sb.append(")");
				break;
			case MREAL :
				sb.append("(");
				sb.append("&gt;&gt;");
				sb.append(",");
				sb.append(getStepTypeStart(step.getType()));
				sb.append(getNodeAbbrv(step.getModelNode()));
				sb.append(getStepTypeEnd(step.getType()));
				sb.append(")");
				break;
			case MINVI :
				sb.append("(");
				sb.append("&gt;&gt;");
				sb.append(",");
				sb.append(getStepTypeStart(step.getType()));
				sb.append(getNodeAbbrv(step.getModelNode()));
				sb.append(getStepTypeEnd(step.getType()));
				sb.append(")");
				break;
			default :
				sb.append("(");
				sb.append(getStepTypeStart(step.getType()));
				sb.append(getNodeAbbrv(step.getModelNode()));
				sb.append(getStepTypeEnd(step.getType()));
				sb.append(",");
				sb.append(getStepTypeStart(step.getType()));
				sb.append(getNodeAbbrv(step.getModelNode()));
				sb.append(getStepTypeEnd(step.getType()));
				sb.append(")");
				break;
		}
		return sb.toString();

	}

	private String buildAlignmentString(ControlFlowAlignmentResult result) {
		StringBuilder sb = new StringBuilder("(");
		Iterator<Step> stepIter = result.getSteps().iterator();
		while (stepIter.hasNext()) {
			Step step = stepIter.next();
			sb.append(getStepTypeStart(step.getType()));
			if (step.getType() == StepTypes.L) {
				sb.append(getNodeAbbrv(step.getLogNode()));
			} else {
				sb.append(getNodeAbbrv(step.getModelNode()));
			}
			sb.append(getStepTypeEnd(step.getType()));
			if (stepIter.hasNext()) {
				sb.append(',');
			}
		}
		sb.append(")");
		return sb.toString();
	}

	private String getStepTypeStart(StepTypes stepType) {
		switch (stepType) {
			case L :
				return "<FONT COLOR=\"gold2\">";
			case MREAL :
				return "<FONT COLOR=\"magenta2\">";
			case MINVI :
				return "<FONT COLOR=\"black\">";
			default :
				return "<FONT COLOR=\"limegreen\">";
		}
	}

	private String getStepTypeEnd(StepTypes stepType) {
		return "</FONT>";
	}

	private String getNodeAbbrv(Object node) {
		if (node instanceof PNWDTransition) {
			if (((PNWDTransition) node).isInvisible()) {
				String nodeStr = node.toString();
				return "<B>t<SUB>" + nodeStr.substring(nodeStr.length() - 1, nodeStr.length()) + "</SUB></B>";
			} else {
				return "<B>" + node.toString().substring(0, 1) + "</B>";
			}
		} else if (node instanceof XEvent) { 
			return "<B>" + delegate.getClassOf((XEvent) node).toString().substring(0, 1) + "</B>";
		} else {
			return "<B>" + node.toString().substring(0, 1) + "</B>";
		}
	}

	@Override
	public void nodeVisited(Record node) {
		ControlFlowAlignmentResult cfAlg = calcCFAlignmentResult(node);
		try {
			writer.write("n" + stateToID(node)
					+ " [label=<"
					//+ "n" + stateToID(node) + " " 
					+ "[" + (++visitCounter) + "]"
					//+ (node.getPredecessor() != null ? "<BR/>" + buildAlignmentString(cfAlg) : "()") 
					//+ "<BR/>" 					
					+ node.getCostSoFar() 
					+ "<BR/>" 
					+ node.getEstimatedRemainingCost()
					+ ">,style=filled,color=lightgray];");
		} catch (Exception e) {

		}
	}

	@Override
	public void edgeTraversed(Record from, Record to) {
		ControlFlowAlignmentResult cfAlg = calcCFAlignmentResult(to);
		try {
			String parameters;
			if (to.getModelMove() == AStarThread.NOMOVE) {
				parameters = "color=gold2";
			} else if (to.getMovedEvent() == AStarThread.NOMOVE) {
				parameters = "color=magenta2";
			} else {
				parameters = "color=limegreen";
			}

			writer.write("n"
					+ stateToID(from)
					+ " -> n"
					+ stateToID(to)
					+ " [penwidth=2.0,label=<"
					+ (to.getCostSoFar() - from.getCostSoFar())
					+ buildLastStepString(cfAlg)
					+ ">,"
					+ (to.getEstimatedRemainingCost() == ObservableAStarThread.ESTIMATEIRRELEVANT ? "arrowhead=tee,"
							: "") + parameters + "];");
			if (to.getEstimatedRemainingCost() != ObservableAStarThread.ESTIMATEIRRELEVANT) {
				writer.write("n" + stateToID(to) + " [label=<"
						//+ "n" + stateToID(to) 
						//+ "<BR/>" 
						+ to.getCostSoFar()
						+ "<BR/>" + to.getEstimatedRemainingCost() + ">];");
			}
		} catch (Exception e) {
		}
	}
	
	private long index = 1;
	
	private String stateToID(Record record) {
		long currentIndex = stateMap.getLong(record);
		if (currentIndex != stateMap.defaultReturnValue()) {
			return String.valueOf(currentIndex);			
		} else {
			currentIndex = index++;
			stateMap.put((DRecord) record, currentIndex);
			return String.valueOf(currentIndex);
		}
	}

	@Override
	public void initialNodeCreated(Record node) {
		try {

			visitCounter = 0;
			finalNodeCounter = 0;
			outputFile.createNewFile();
			OutputStream stream = new BufferedOutputStream(new FileOutputStream(outputFile));
			this.writer = new OutputStreamWriter(stream);
			writer.append("digraph G { ");
			writer.write("n" + stateToID(node) + " [peripheries=2];");
		} catch (IOException e) {
			this.writer = null;
		}
	}

	@Override
	public void finalNodeFound(Record node) {
		finalNodeCounter++;
		try {
			writer.write("n" + stateToID(node) + " [peripheries=2];");
			addPathToRoot(node);
			close();
		} catch (Exception e) {
		}
	}

	public void close() {
		try {
			if (writer != null) {
				writer.write("}");
				writer.close();
			}
		} catch (Exception e) {
		}
	}

	protected void addPathToRoot(Record node) throws IOException {
		while (node.getPredecessor() != null) {
			writer.write("n" + stateToID(node.getPredecessor()) + " -> n" + stateToID(node)
					+ " [weight=1,penwidth=5.0,style=dashed,color=red,arrowhead=none,arrowtail=none];");
			node = node.getPredecessor();
		}
	}

	@Override
	public void stoppedUnreliablyAt(Record node) {
		if (finalNodeCounter > 0) {
			return;
		}
		finalNodeCounter++;
		try {
			if (node != null) {
				writer.write("n" + stateToID(node) + " [peripheries=2,shape=diamond];");
				addPathToRoot(node);
			}
			writer.write("}");
			writer.close();
		} catch (IOException e) {
		}
	}

	@Override
	public void estimateComputed(Head head) {
		// skip
	}

}
